/* morseteach
 * A Unix implementation of Ward Cunningham's automatic morse teacher.
 * pesco DC4HK 2021
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>	/* timespecsub */
#include <ctype.h>
#include <math.h>	/* sqrt, sin, exp */
#include <assert.h>
#include <unistd.h>	/* sleep */
#include <termios.h>
#include <err.h>

/* available symbol sequences */
const char AB_ORIG[]	= "Q7ZG098O1JPWLRAM6BXDYCKN23FU45VHSITE";
const char AB_FLTK[]	= "Q7ZG098O1JPW.LRAM6B/XD=YCKN23?FU45VHSITE";
const char AB_FULL[]	= "Q7ZG098O1JPWLRAM6BXDYCKN23FU45VHSITE./=?,+-:;\"'()@";
const char AB_LCWO[]	= "KMURESNAPTLWI.JZ=FOY,VG5/Q92H38B?47C1D60X";
const char AB_LOCKD[]	= "EOAIUYZQJXKVBPGWFCLDMHRSNT0156273849?/.,";
const char AB_DJ4UF[]	= "ESNOTQ5RU9=CD0AL8IX4.M7Y1H6?J3G2VK,ZFBPW/";
const char AB_G4FON[]	= "KMRSUAPTLOWI.NJEF0YV,G5/Q9ZH38B?427C1D6X=";

struct alphabet { const char *name; const char *syms; } alphabets[] = {
	{ "orig", AB_ORIG },	/* sequence used in original morse machine */
	{ "fltk", AB_FLTK },	/* with punctuation as found in FLTK version */
	{ "full", AB_FULL },	/* with all (official) punctuation at the end */
	{ "lcwo", AB_LCWO },	/* from "Learn CW Online" by DJ1YFK */
	{ "lockd", AB_LOCKD },	/* from "Lockdown Morse" course by M0PTO */
	{ "dj4uf", AB_DJ4UF },	/* from "Morselehrgang" by DJ4UF */
	{ "g4fon", AB_G4FON },	/* from "Koch CW Trainer" by G4FON */
	{ NULL, NULL }
};

const char *alphabet = AB_ORIG;	/* the symbol sequence being used */
int nstart = 2;		/* number of active symbols to start with */
int diagnostic = 0;	/* level of diagnostic output */
double threshold = 0.3;	/* maximum error rate before adding a character */
double initerr = 0.5;	/* initial error estimate */
double newerr = 1;	/* initial error estimate of a newly added symbol */
double initrsp = 1;	/* initial response time estimate */
double initvar = 0;	/* initial variance estimate */
double damping = 0.875;	/* (7/8) weighting factor for past estimate value */

/* teacher state */
int ntotal;		/* number of symbols in alphabet */
int nactive;		/* number of characters currently being trained */
double *error;		/* estimated error rate for each character */
double overall;		/* estimated overall error rate */
double avgtime;		/* estimated average response time */
double vartime;		/* estimated variance in response time */
struct timespec tbegin;	/* when the session started */
struct timespec tpause;	/* time spent paused */

/* other globals */
FILE *fout, *fin;	/* our pipe to 'morse -s | morseplay' */

void
estimate(double *var, double x)
{
	*var = *var * damping + x * (1 - damping);
}

void
printscores(void)
{
	struct timespec t, trun;
	int i, j;

	/* calculate session run time */
	clock_gettime(CLOCK_MONOTONIC, &t);
	timespecsub(&t, &tbegin, &trun);
	timespecsub(&trun, &tpause, &trun);

	/* bargraph */
	for (j = 3; j > 0; j--) {
		for (i = 0; i < nactive; i++)
			if (error[i] >= j * 0.3)
				putchar('|');
			else if (error[i] >= j * 0.3 - 0.1)
				putchar(':');
			else if (error[i] >= j * 0.3 - 0.2)
				putchar('.');
			else
				putchar(' ');
		putchar('\n');
	}
	for (i = 0; i < nactive; i++)
		putchar(alphabet[i]);
	putchar('\n');

	printf("average response time %.0f ms (%.1f wpm), ",
	    avgtime * 1000, 3.6 / avgtime);
	printf("overall error rate %.0f%%\n", overall * 100);

	printf("training level %d. session duration:", nactive);
	if (trun.tv_sec >= 3600) {
		printf(" %lu h", (long)trun.tv_sec / 3600);
		trun.tv_sec %= 3600;
	}
	if (trun.tv_sec >= 60)
		printf(" %lu min", (long)trun.tv_sec / 60);
	else
		printf(" %lu s", (long)trun.tv_sec);
	putchar('\n');
}

void
play(char c)
{
	fputc(c, fout);
	while (fgetc(fin) != '\n') ;	/* wait for sound to finish */
}

const char *usage = "morseteach [-h] [level]";

int
main(int argc, char **argv)
{
	struct termios t, bak;
	struct timespec t0, t1, delta, tend;
	double l, x;
	double result;
	double taken, sigma;
	int ch, input;
	int i;
	int tries;
	char *val;
	struct alphabet *ab;

	/* environment variables */
	if ((val = getenv("MORSE_LEVEL")) != NULL) {
		if ((i = atoi(val)) > 0)
			nstart = i;
		else
			warnx("MORSE_LEVEL invalid, defaulting to %d", nstart);
	}
	if ((val = getenv("MORSE_ALPHABET")) != NULL) {
		/* look for an alphabet with the given name */
		for (ab = alphabets; ab->name != NULL; ab++)
			if (strcmp(ab->name, val) == 0)
				break;

		/* if val is none of the defined names, use it literally */
		if (ab->name == NULL)
			alphabet = val;
		else
			alphabet = ab->syms;
	}

	/* command line arguments */
	while ((ch = getopt(argc, argv, "hD")) != -1) {
		switch (ch) {
		case 'D':
			diagnostic++;
			break;
		case 'h':
			puts(usage);
			return 0;
		default:
			puts(usage);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		nstart = atoi(argv[0]);
		if (nstart < 1)
			errx(1, "%s: invalid level", argv[0]);
	}

	/* disable line "canonicalization", buffering, and echo on stdin */
	if (tcgetattr(0, &t) != 0)
		err(1, "tcgetattr");
	bak = t;
	t.c_lflag &= ~(ICANON | ECHO);
	t.c_cc[VMIN] = 1;	/* wait for a single character */
	t.c_cc[VTIME] = 0;	/* do not time out */
	if (tcsetattr(0, TCSANOW, &t) != 0)
		err(1, "tcsetattr");

	/* disable buffering on stdout */
	setvbuf(stdout, NULL, _IONBF, 0);

	/* open an unbuffered channel to the morse player */
	if ((fin = fdopen(3, "r")) == NULL)
		err(1, "fdopen: 3");
	if ((fout = fdopen(4, "w")) == NULL)
		err(1, "fdopen: 4");
	setvbuf(fout, NULL, _IONBF, 0);

	/* initialize teacher state */
	ntotal = strlen(alphabet);
	if (nstart > ntotal)
		nstart = ntotal;
	nactive = nstart;
	if((error = malloc(sizeof(double) * ntotal)) == NULL)
		err(1, "malloc");
	for (i = 0; i < ntotal; i++)
		error[i] = initerr;
	overall = initerr;
	avgtime = initrsp;
	vartime = initvar;
	srand(time(NULL));

	puts("Type what you hear!\n"
	    "Don't worry about mistakes, but try to be accurate rather than fast.\n"
	    "Press Space for a hint, Return to pause, ^D to quit.");
	fputs("Training:", stdout);
	for (i = 0; i < nactive; i++)
		printf(" %c", toupper(alphabet[i]));
	putchar('\n');
	fputs(" - press Return to begin - ", stdout);
	while ((input = getchar()) != '\n' && input != 4) ;
	if (input == 4)
		puts("^D");
	else if (input == '\n') {
		play('\a');	/* <KA> attention / start of transmission */
		play('\n');	/* inter-word pause */
		puts("\r                          ");
	}

	clock_gettime(CLOCK_MONOTONIC, &tbegin);
	while (!feof(stdin) && input != 4) {
		/*
		 * pick a random character from the active training set,
		 * distributed according to the error rates.
		 */
		l = 0;
		for (i = 0; i < nactive; i++)
			l += error[i];
		x = l * rand() / RAND_MAX;
		assert(nactive > 1);
		for (i = 0; i < nactive - 1; i++) {
			x -= error[i];
			if (x < 0)
				break;
		}
		assert (i < nactive);
		if (nactive > 2 && alphabet[i] == ch)	/* avoid direct repeats */
			continue;
		ch = alphabet[i];

		/*
		 * play the symbol.
		 * new/unknown symbols are emphasized by a short pause and
		 * automatically shown after playback.
		 */
		if (error[i] > 0.9)
			play(' ');
		play(ch);
		clock_gettime(CLOCK_MONOTONIC, &t0);
		if (error[i] > 0.9)
			printf("%c\b", ch);

		tries = 0;
		result = 1;	/* failure */
		while (!feof(stdin)) {
			/* wait for the student to answer */
			input = getchar();

			/* measure response time */
			clock_gettime(CLOCK_MONOTONIC, &t1);
			timespecsub(&t1, &t0, &delta);
			taken = delta.tv_sec + delta.tv_nsec / 10000000000.0;	/* [s] */
			sigma = sqrt(vartime);

			/* if return pressed, pause and show scores */
			if (input == '\n') {
				clock_gettime(CLOCK_MONOTONIC, &t0);
				putchar('\n');
				printscores();
				fputs("- pause (press Return to continue, ^D to quit) - ", stdout);
				play('&');		/* <AS> wait */
				while ((input = getchar()) != '\n' && input != 4) ;
				clock_gettime(CLOCK_MONOTONIC, &t1);
				timespecsub(&t1, &t0, &delta);
				timespecadd(&tpause, &delta, &tpause);
				if (input == 4) {
					puts("^D");
					play('+');	/* <AR> end of transmission */
					play('\n');
					break;
				} else {
					play('\a');	/* <KA> attention / start of transmission */
					play('\n');	/* inter-word pause */
					puts("\r                                                ");
					play(ch);
					clock_gettime(CLOCK_MONOTONIC, &t0);
					continue;
				}
			}

			/* if ^D pressed, then quit */
			else if (input == 4) {
				puts("^D");
				play('+');	/* <AR> end of transmission */
				play('\n');
				printscores();
				break;
			}

			tries++;

			/* if space pressed, resound and show hint */
			if (input == ' ') {
				printf("%c\b", ch);
				play(ch);
				printf(" \b");
				continue;
			}

			/* if answered correctly, display input and record response time */
			else if (toupper(input) == toupper(ch)) {
				if (taken <= avgtime + sigma)
					result = 0;	/* success */

				if (diagnostic >= 1 && result != 0)
					printf("\e[1m%c\e[0m", input);	/* bold */
				else
					putchar(input);
				if (diagnostic >= 2)
					printf("\ntook %.2f s, avg %.2f s (+/- %.2f s), %s. ",
					    taken, avgtime, sigma, result > 0 ? "fail" : "pass");

				estimate(&vartime, (avgtime - taken) * (avgtime - taken));
				if (taken > avgtime + sigma)
					taken = avgtime + sigma;
				else if (taken < avgtime - sigma)
					taken = avgtime - sigma;
				estimate(&avgtime, taken);
				break;
			} else if (taken > avgtime + sigma) {
				/* resound the symbol */
				play(ch);
			}
		}
		if (input == 4)	/* ^D */
			break;

		estimate(&overall, result);
		estimate(&error[i], result);
		if (diagnostic >= 2)
			printf("error rate for %c %2.0f%%, overall %2.0f%%\n",
			    ch, error[i] * 100, overall * 100);

		/* add a new character if all error rates look good (<= threshold) */
		if (nactive < ntotal && overall <= threshold) {
			for (i = 0; i < nactive; i++)
				if (error[i] > threshold)
					break;
			if (i == nactive) {
				if (diagnostic >= 3)
					printf("all error rates <= %.0f%% - new character: %c\n",
					    threshold * 100, alphabet[nactive]);
				error[nactive] = newerr;
				nactive++;
			}
		}
	}
	clock_gettime(CLOCK_MONOTONIC, &tend);
	fclose(fout);			/* morse plays <SK> */
	while (fgetc(fin) != EOF) ;	/* drain the pipe */

	/* restore terminal mode */
	if (tcsetattr(0, TCSANOW, &bak) != 0)
		err(1, "tcsetattr");

	return 0;
}
